
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/fs.h>
#include <linux/hdreg.h>
#include <assert.h>
#include <string.h>
#include <errno.h>

#include "dvh.h"

#define SECTORS_PER_TRACK	32
#define BYTES_PER_SECTOR	512

#define __u16	unsigned short
#define __u32	unsigned int

#define swab16(x) \
({ \
	__u16 __x = (x); \
	((__u16)( \
		(((__u16)(__x) & (__u16)0x00ffU) << 8) | \
		(((__u16)(__x) & (__u16)0xff00U) >> 8) )); \
})

#define swab32(x) \
({ \
	__u32 __x = (x); \
	((__u32)( \
		(((__u32)(__x) & (__u32)0x000000ffUL) << 24) | \
		(((__u32)(__x) & (__u32)0x0000ff00UL) <<  8) | \
		(((__u32)(__x) & (__u32)0x00ff0000UL) >>  8) | \
		(((__u32)(__x) & (__u32)0xff000000UL) >> 24) )); \
})

#if __BYTE_ORDER == __LITTLE_ENDIAN
#define be32_to_cpu(x) swab32((x))
#define be16_to_cpu(x) swab16((x))
#define cpu_to_be32(x) swab32((x))
#define cpu_to_be16(x) swab16((x))
#else
#define be32_to_cpu(x) ((__u32)(x))
#define be16_to_cpu(x) ((__u16)(x))
#define cpu_to_be32(x) ((__u32)(x))
#define cpu_to_be16(x) ((__u16)(x))
#endif


__u32 vh_calc_checksum(struct volume_header *vh) {
	__u32	oldchecksum,
		newsum=0,
		*buffer=(__u32 *) vh;
	int	i;

	oldchecksum=vh->vh_csum;
	vh->vh_csum=0;

	for(i=0;i<sizeof(struct volume_header)/sizeof(__u32);i++) {
		newsum=newsum-be32_to_cpu(buffer[i]);
	}

	vh->vh_csum=oldchecksum;
	return(newsum);
}

int main(int argc, char **argv) {
	int			fd,
				i,
				block,
				blocksize;
	struct stat		st;
	struct statfs		stfs;
	struct volume_header	vh;

	assert(argv[1] != NULL);

	assert(fd=open(argv[1], O_RDONLY));

	/* Get blocksize of filesystem - ISO9660 -> 2048 */
	assert(ioctl(fd, FIGETBSZ, &blocksize) == 0);

	/* Get filesize */
	assert(!fstat(fd, &st));
	/* Get filesystem size to later calc virtual volume header
	 * size */
	assert(!fstatfs(fd, &stfs));

	printf("File: %s Size: %d Blocksize: %d\n", 
		argv[1], st.st_size, blocksize);

	/* We only get the first block - iso9660 files are contigues
	 * and for the sgi volume header we do not support fragments */
	block=0;
	if (ioctl(fd, FIBMAP, &block)) {
		printf("FIBMAP ioctl failed - errno: %s\n",
				strerror(errno));
		exit(1);
	}

	close(fd);

	memset(&vh, 0, sizeof(struct volume_header));

	vh.vh_magic = cpu_to_be32(VHMAGIC);

	/* Values from an IRIX cd */
	vh.vh_dp.dp_secbytes=cpu_to_be16(BYTES_PER_SECTOR);
	vh.vh_dp.dp_secs=cpu_to_be16(SECTORS_PER_TRACK);
	vh.vh_dp.dp_flags=cpu_to_be32(DP_RESEEK|DP_IGNOREERRORS|DP_TRKFWD);

	vh.vh_dp.dp_cyls=cpu_to_be16((stfs.f_blocks * blocksize) /
		(SECTORS_PER_TRACK*BYTES_PER_SECTOR));

	/* Set directory entry name */
	strcpy(vh.vh_vd[0].vd_name, "linux");
	vh.vh_vd[0].vd_lbn=cpu_to_be32((block * blocksize) / 512);
	vh.vh_vd[0].vd_nbytes=cpu_to_be32(st.st_size);

	/* Create volume header partition on WHOLE cd iso */
	vh.vh_pt[8].pt_nblks=cpu_to_be32((stfs.f_blocks * blocksize) /
		BYTES_PER_SECTOR);

	/* Create checksum */
	vh.vh_csum=cpu_to_be32(vh_calc_checksum(&vh));

	assert((fd=open("volhdr", O_RDWR|O_EXCL|O_CREAT)));
	write(fd, &vh, sizeof(struct volume_header));
	close(fd);

}
